//  $Id: World.cc,v 1.58 1999/12/22 23:58:19 grumbel Exp $
//
//  Pingus - A free Lemmings clone
//  Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <vector>
#include <cstdio>
#include <ctime>
#include <algorithm>
#include <functional>

#include "PinguBmpMap.hh"
#include "PingusSpotMap.hh"
#include "PinguRandomMap.hh"

#include "PingusError.hh"
#include "algo.hh"
#include "PingusSound.hh"
#include "globals.hh"
#include "Background.hh"
#include "World.hh"
#include "traps/traps.hh"
#include "Result.hh"
#include "Liquid.hh"
#include "actions/pingu_actions.hh"
#include "actions/bomber.hh"
#include "entrances/entrances.hh"
#include "FVec.hh"

// Workaround for a VC bug
#ifdef WIN32
#  define for if(0);else for
#endif /* WIN32 */

// Structure for the sorting algorithm (stable_sort)
struct WorldObj_less : public binary_function<WorldObj*, WorldObj*, bool>
{
  bool operator() (WorldObj* a, WorldObj* b) const 
    {
      return (*a) < (*b);
    }
};

World::World()
{
  released_pingus = 0;
  saved_pingus = 0;

  exit_world = false;
}

World::World(PLF* plf)
{ 
  init(plf);
}

World::~World()
{
  // Bug: This crashes, if pingus is exited
  delete map;
  delete background;

  for(vector<Exit*>::size_type i = 0; i < exits.size(); ++i) 
    delete exits[i];
  
  for(vector<Entrance*>::size_type i = 0; i < entrance.size(); ++i) 
    delete entrance[i];
 
  for(vector<Trap*>::size_type i = 0; i < traps.size(); ++i)
    delete traps[i];

  for(vector<Hotspot*>::size_type i = 0; i < hotspot.size(); ++i)
    delete hotspot[i];
  
  for(vector<Liquid*>::size_type i = 0; i < liquid.size(); ++i) {
    delete liquid[i];
  }
}

// Merge the different layers on the screen together
void 
World::draw_offset(int x_of, int y_of, float s)
{
  background->draw_offset(x_of, y_of, s);

  for(vector<WorldObj*>::iterator obj = world_obj.begin(); obj != world_obj.end(); obj++)
    {
      (*obj)->draw_offset(x_of, y_of, s);
    }
  
  particle.draw_offset(x_of, y_of, s);
}

void 
World::let_move()
{
  if (!exit_world && (allowed_pingus == released_pingus || do_armageddon)
      && pingus_out == 0) 
    {
      cout << "World finished" << endl;
      exit_world = true;
      exit_time = GameTime::get_time() + 75;
    }

  if (do_armageddon && armageddon_count != pingus.end()) 
    {
      if (armageddon_count == 0) {
	armageddon_count = pingus.begin();
      } else {
	int i = 0;
	for(PinguIter pingu = armageddon_count; i < 5; ++pingu, ++i) {
	  armageddon_count = pingu;	
	  if (armageddon_count == pingus.end())
	    break;
	  if ((*pingu)->get_status() != dead)
	    (*pingu)->set_action(new Bomber);
	  else
	    --i;
	}
      }
    }
  
  // Create new pingus, if enough time is passed
  if (!do_armageddon) 
    {
      for(vector<Entrance*>::iterator i = entrance.begin(); i != entrance.end(); ++i) 
	{
	  if (pingus.size() < allowed_pingus
              && (*i)->pingu_ready())
	    {
	      pingus.push_back((*i)->get_pingu());
	      ++released_pingus;
	    }
	}
    }
  
  // Let all pingus move and
  // Let the pingus catch each other and
  // Let the traps catch the pingus and
  // Let the exit catch the pingus
  pingus_out = pingus.size();
  saved_pingus = 0;

  for(PinguIter pingu = pingus.begin(); pingu != pingus.end(); ++pingu)
    {
      (*pingu)->let_move();
      
      if ((*pingu)->get_status() == dead
	  || (*pingu)->get_status() == exited) {
	if ((*pingu)->get_status() == exited)
	  ++saved_pingus;
	--pingus_out;
	continue;
      }
    
      if ((*pingu)->need_catch()) {
	for(PinguIter i = pingus.begin(); i != pingus.end(); i++) {
	  (*pingu)->catch_pingu(*i);
	}
      }
    
      for(vector<Trap*>::iterator obj = traps.begin(); obj != traps.end(); obj++)
	(*obj)->catch_pingu(*pingu);
      
      for(vector<Exit*>::iterator obj = exits.begin(); obj != exits.end(); obj++) 
	(*obj)->catch_pingu(*pingu);
    }

  for(vector<trap_data>::size_type i=0; i < traps.size(); ++i)
    traps[i]->let_move();
  
  for(vector<entrance_data>::size_type i2=0; i2 < entrance.size(); ++i2) 
    entrance[i2]->let_move();
  
  particle.let_move();
  background->let_move();
  
  // Clear the explosion force list
  if (force_vectors)
    ForcesHolder::clear_explo_list();
}

void 
World::init(PLF* plf_data)
{
  plf = plf_data;
  do_armageddon = false;
  armageddon_count = 0;
  allowed_pingus = plf->get_pingus();
  released_pingus = 0;
  pingus_out = 0;

  exit_time = plf->get_time();
  cout << "Exit Time: " << exit_time << endl;

  if (verbose)
    cout << "World init" << endl;

  init_map();
  init_background();
  init_worldobjs();
}

void
World::init_map()
{
  // load the foreground map
  switch (plf->map_type()) 
    {
    case SPOT:
      map = new PingusSpotMap(plf);
      break;
    case BMP:
    case RANDOM:
    default:
      throw PingusError("Error in PLF file: Undef or unsupported Maptype");
      break;
    }
  
  //  if (plf->get_mapfile().type == ResDescriptor::AUTO) 
  //    {
      if (verbose) cout << "Using AutoColMap" << endl;
      colmap = map->get_colmap();

      /*    } 
       else
    {
      if (verbose) cout << "Using StaticColMap" << endl;
      colmap = new ColMap;
      colmap->load(plf->get_mapfile());
    }
      */
  if (verbose) 
    {
      cout << "Loaded Colmap" << endl;
      cout << "Loaded pingu map" << endl;
    }
}

void 
World::init_background()
{
  // load the background map
  background = new Background(plf->get_bg());

  if (verbose)
    cout << "Loaded Background" << endl;
}

void
World::init_worldobjs()
{
  vector<exit_data>     exit_d     = plf->get_exit();
  vector<entrance_data> entrance_d = plf->get_entrance();
  vector<trap_data>     trap_d     = plf->get_traps();
  vector<hotspot_data>  hspot_d    = plf->get_hotspot();
  vector<liquid_data>   liquid_d   = plf->get_liquids();

  // Creating Exit and Entrance
  for(vector<exit_data>::iterator i = exit_d.begin(); i != exit_d.end(); i++) 
    exits.push_back(new Exit(*i));
  
  for(vector<entrance_data>::size_type i = 0; i < entrance_d.size(); ++i) 
    entrance.push_back(get_entrance(entrance_d[i]));

  for(vector<trap_data>::size_type i=0; i < trap_d.size(); ++i)
    traps.push_back(get_trap(trap_d[i]));

  // Creating the foreground and background hotspots

  for(vector<hotspot_data>::size_type i = 0; i < hspot_d.size(); ++i)
    hotspot.push_back(new Hotspot(hspot_d[i]));
  
  for(vector<liquid_data>::size_type i=0; i < liquid_d.size(); i++) 
    liquid.push_back(new Liquid(liquid_d[i]));
  
  Pingu::set_colmap(colmap);
  Pingu::set_map(map);
  Pingu::SetParticleHolder(&particle);
  Trap::SetParticleHolder(&particle);
  Entrance::SetParticleHolder(&particle);

  world_obj.push_back(map);

  // Push all objects to world_obj vector
  for(vector<hotspot_data>::size_type i = 0; i < hotspot.size(); i++)
    world_obj.push_back(hotspot[i]);

  for(vector<exit_data>::size_type i=0; i < exits.size(); i++)
    world_obj.push_back(exits[i]);
  
  for(vector<entrance_data>::size_type i=0; i < entrance.size(); ++i)
    world_obj.push_back(entrance[i]);

  for(vector<liquid_data>::size_type i=0; i < liquid_d.size(); ++i)
    world_obj.push_back(liquid[i]);

  for(vector<trap_data>::size_type i=0; i < traps.size(); ++i) 
    world_obj.push_back(traps[i]);

  world_obj.push_back(&pingus);

  // After all objects are in world_obj, sort them to there z_pos
  stable_sort(world_obj.begin(), world_obj.end(), WorldObj_less());
  
  if (verbose) 
    cout << "Finished Word init" << endl;
  
  if (sound_enabled) PingusSound::play(PingusSound::LETSGO);

  // Setup the gravity force
  if (force_vectors)
    {
      // Clear all old forces
      ForcesHolder::clear_all_forces();
      ForcesHolder::add_force(new GravityForce(grav));
    }
}

PinguHolder*
World::get_pingu_p(void)
{
  return &pingus;
}

int
World::get_width(void)
{
  assert(map);
  return map->get_width();  
}

int
World::get_height(void)
{
  assert(map);
  return map->get_height();
}

int
World::get_time(void)
{
  if (exit_time == 0) 
    {
      return GameTime::get_time();
    }
  else
    {
      return exit_time - GameTime::get_time();
    }
}

void
World::write_result(void)
{
  result.killed  = 0;
  result.saved   = 0;
  result.unknown = 0;
  result.total   = allowed_pingus;
  result.time    = GameTime::get_time();
  
  for(PinguIter pingu = pingus.begin(); pingu != pingus.end(); pingu++)
    {
      switch((*pingu)->get_status()) 
	{
	case dead:
	  result.killed++;
	  break;
	case exited:
	  result.saved++;
	  break;
	default:
	  result.unknown++;
	}
    }
}

Result
World::get_result()
{
  write_result();
  return result;
}

void 
World::armageddon(void)
{
  cout << "Doing armo" << endl;
  do_armageddon = true;
}

bool
World::is_finished(void)
{
  // Return true if the world is finished and some time has passed
  if (exit_world && exit_time < GameTime::get_time()) 
    {
      return true;
    } 
  else if (exit_world)
    {
      cout << "The world will exit in: " << exit_time << " : " << GameTime::get_time() << endl;
      return false;
    } 
  else 
    {
      return false;
    }
}

void
World::print_status(void)
{
  cout << "Released Pingus: " << released_pingus << endl;
  cout << "Allowed Pingus:  " << allowed_pingus << endl;
  cout << "Pingus Out:      " << pingus_out << endl;
}

/* EOF */
